/* -*- c -*- --------------------------------------------------------------- *
 *
 * linux/fs/autofs/symlink.c
 *
 *  Copyright 1997-1998 Transmeta Corporation -- All Rights Reserved
 *
 * This file is part of the Linux kernel and is made available under
 * the terms of the GNU General Public License, version 2, or at your
 * option, any later version, incorporated herein by reference.
 *
 * ------------------------------------------------------------------------- */

#include "autofs_i.h"

static inline int vfs_readlink(struct dentry *dentry, char *buffer, int buflen, 
			       const char *link)
{
	int len;

	len = PTR_ERR(link);
	if (IS_ERR(link))
		goto out;

	len = strlen(link);
	if (len > (unsigned) buflen)
		len = buflen;
	if (copy_to_user(buffer, link, len))
		len = -EFAULT;
out:
	return len;
}

static int autofs4_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	struct autofs_info *ino = autofs4_dentry_ino(dentry);

	return vfs_readlink(dentry, buffer, buflen, ino->u.symlink);
}

static struct dentry * autofs4_follow_link(struct dentry *dentry,
					  struct dentry *base,
					  unsigned int flags)
{
	struct autofs_info *ino = autofs4_dentry_ino(dentry);

	return lookup_dentry(ino->u.symlink, base, flags);
}

struct inode_operations autofs4_symlink_inode_operations = {
	readlink:	autofs4_readlink,
	follow_link:	autofs4_follow_link
};
